"""A sound editor.

Scalpel is a sound editor. It works only on Linux at this point, with
a GTK frontend and an ALSA backend.

Changelog for this release at: http://scalpelsound.online.fr/?p=388

"""
from constants import __version__
